/*
 *  arch/ppc/8xx_io/snomdisp.c
 *
 *  Copyright (C) 2000, snom technology AG
 *
 *  This code is for the LCD module connected to the SNOM board.
 *  It only implements the write system call and expects 1024 bytes
 *  to be written. It dumps these onto the display.
 *
 */

/***********************************************************************

 This driver is used to connect the PG 12864 J to the SNOM board and 
 thus be able to configure it. 

 The data lines are expect to be connected as follows:

 1  Vdd
 2  Vss
 3  Vo
 4  DB0  - PB14 P6.7,  PM3.97       - 1
 5  DB1  - PB15 P6.9,  PM3.95       - 2
 6  DB2  - PB16 P6.11, PM3.93       - 3
 7  DB3  - PB17 P6.13, PM3.75       - 4
 8  DB4  - PB18 P6.15, PM3.76       - 5
 9  DB5  - PB22 P6.23, PM3.44       - 6
 10 DB6  - PB23 P6.25, PM3.40       - 7
 11 DB7  - PB26 P6.31, PM3.54       - 8
 12 CS1  - PB27 P6.33, PM3.50       
 13 CS2  - PB28 P6.35, PM3.51
 14 RST  - PB29 P6.37, PM3.37
 15 RW   - PB30 P6.10, PM3.35
 16 D/I  - PB31 P6.12, PM3.33
 17 E    - PC4  P6.14, PM3.98
 18 Vss   
 19 A 
 20 K 

The SNOM CSR bit 4 controls whether the backlight is on or off.

************************************************************************/



#include <linux/config.h>
#include <linux/types.h>

#ifdef CONFIG_SNOM_DISP

int global_serial_num;

#include <asm/port.h>
#include <asm/mpc8xx.h>

#include "commproc.h"

/* Constants */

#define DELAY		for(j=0;j<50;j++);

/* The data bus and these lines are located on port B. */

#define DB0_B             (1<<17)
#define DB1_B             (1<<16)
#define DB2_B             (1<<15)
#define DB3_B             (1<<14)
#define DB4_B             (1<<13)
#define DB5_B             (1<<9)
#define DB6_B             (1<<8)
#define DB7_B             (1<<5)

#define CS1_B             (1<<4)
#define CS2_B             (1<<3)
#define RST_B             (1<<2)
#define RW_B              (1<<1)
#define DI_B              (1<<0)

/* This is on port C. */

#define EN_C              (1<<11)

/* Virtual values */

#define CS1             (1<<0)
#define CS2             (1<<1)
#define RST             (1<<2)
#define RW              (1<<3)
#define DI              (1<<4)
#define EN              (1<<5)

/* Command values */

#define CMD_DISP_OFF    0x3e
#define CMD_DISP_ON     0x3f
#define CMD_SET_X       0xB8
#define CMD_SET_Y       0x40
#define CMD_SET_Z       0xC0

/* Debug */

#if 0
#define DEB(x)        printf(x);
#define DEB2(x,y)     printf(x,y);
#define DEB3(x,y,z)   printf(x,y,z);
#define DEB4(x,y,z,a) printf(x,y,z,a);
#else
#define DEB(x)
#define DEB2(x,y)
#define DEB3(x,y,z)
#define DEB4(x,y,z,a)
#endif

static volatile immap_t	  *immap = (immap_t *)IMAP_ADDR;

/* Write command */
static void write_command(unsigned int chip, unsigned int command)
{
  int port_b_mask = DB0_B | DB1_B | DB2_B | DB3_B | DB4_B | DB5_B | DB6_B | DB7_B | CS1_B | CS2_B | RST_B | RW_B | DI_B;
  volatile unsigned int *pb;
  volatile unsigned short *pc;
  int v, j;

  pb = &(immap->im_cpm.cp_pbdat); 
  pc = &(immap->im_ioport.iop_pcdat); 
  
  /* set the values: */
  v = *pb & ~port_b_mask;

  if(command & (1<<0)) v |= DB0_B;
  if(command & (1<<1)) v |= DB1_B;
  if(command & (1<<2)) v |= DB2_B;
  if(command & (1<<3)) v |= DB3_B;
  if(command & (1<<4)) v |= DB4_B;
  if(command & (1<<5)) v |= DB5_B;
  if(command & (1<<6)) v |= DB6_B;
  if(command & (1<<7)) v |= DB7_B;

  if (chip == 0) v |= CS1_B;
  else v |= CS2_B;

  v |= RST_B;

  *pb = v;
  DELAY;

  /* toggle en line: */
  *pc |= EN_C;
  DELAY;
  *pc &= ~EN_C;
  DELAY;

  *pb = v & ~(CS1_B | CS2_B);
}

/* Write data */
static void write_data(unsigned int chip, unsigned int data)
{
  int port_b_mask = DB0_B | DB1_B | DB2_B | DB3_B | DB4_B | DB5_B | DB6_B | DB7_B | CS1_B | CS2_B | RST_B | RW_B | DI_B;
  volatile unsigned int *pb;
  volatile unsigned short *pc;
  int v, j;

  pb = &(immap->im_cpm.cp_pbdat); 
  pc = &(immap->im_ioport.iop_pcdat); 
  
  /* set the values: */
  v = *pb & ~port_b_mask;

  if(data & (1<<0)) v |= DB0_B;
  if(data & (1<<1)) v |= DB1_B;
  if(data & (1<<2)) v |= DB2_B;
  if(data & (1<<3)) v |= DB3_B;
  if(data & (1<<4)) v |= DB4_B;
  if(data & (1<<5)) v |= DB5_B;
  if(data & (1<<6)) v |= DB6_B;
  if(data & (1<<7)) v |= DB7_B;

  if (chip == 0) v |= CS1_B;
  else v |= CS2_B;

  v |= RST_B | DI_B;

  *pb = v;
  DELAY;

  /* toggle en line: */
  *pc |= EN_C;
  DELAY;
  *pc &= ~EN_C;
  DELAY;

  *pb = v & ~(CS1_B | CS2_B);
}

/* Configuration sequence of display */
static int x, y; 

static void ks0108b_init(void)
{
  int i, j, k;
  volatile unsigned int *pb;
  volatile unsigned short *pc;

  pb = &(immap->im_cpm.cp_pbdat); 
  pc = &(immap->im_ioport.iop_pcdat); 

  /*** Set the ports to the correct mode. */

  set_port_dir(PORT_B, OUTPUT_PINS,
	       DB0_B | DB1_B | DB2_B | DB3_B |
	       DB4_B | DB5_B | DB6_B | DB7_B |
	       CS1_B | CS2_B | RST_B | RW_B  |
	       DI_B);

  set_port_dir(PORT_C, OUTPUT_PINS, EN_C);

  /* reset */
  *pb &= ~RST_B;
  *pc &= ~EN_C;
  DELAY;
  *pb |= RST_B;
  DELAY;

  /* clear the display with zeros ... */
  write_command(0, CMD_SET_X);
  write_command(1, CMD_SET_X);

  write_command(0, CMD_SET_Y);
  write_command(1, CMD_SET_Y);  

  write_command(0, CMD_SET_Z);
  write_command(1, CMD_SET_Z);  

  for(i=0; i<2; i++) {
    for(j=0; j<8; j++) {
      write_command(i, CMD_SET_X | j);
      for(k=0; k<64; k++) {
	write_data(i, 0); 
      }
    }
  }

  x = y = 0;


  /* turn displays on */
  write_command(0, CMD_DISP_ON);
  write_command(1, CMD_DISP_ON);  

  write_command(0, CMD_SET_X);
  write_command(1, CMD_SET_X);

  write_command(0, CMD_SET_Y);
  write_command(1, CMD_SET_Y);  

  write_command(0, CMD_SET_Z);
  write_command(1, CMD_SET_Z);  

}

/* 0 - idle, 1 - got x, 2 - got y, 3 - still need to bytes. */
static int write_state = 0;
static int x = 0, y = 0, len = 0, controller = 0;

int write_snom_display(const char * buf, int count)
{
  int l;

  if(count < 0)
    return -1;

  for(l = 0; l < count; l++) {

    switch(write_state) {
    case 0:
      {
	x = *((const unsigned char *) buf);
	write_state = 1;
	break;
      }
    case 1:
      {
	y = *((const unsigned char *) buf);
	write_state = 2;
	break;
      }
    case 2:
      {
	len = *((const unsigned char *) buf);
	write_state = 3;

	/* setup correct position. */
	
	controller = (x < 64) ? 1 : 0;
	write_command(controller, CMD_SET_X | (y % 8)); 
	write_command(controller, CMD_SET_Y | (x % 64)); 
	break;
      }
    case 3:
      {
	write_data(controller, *buf);
	len--;
	if (len == 0) write_state = 0;
	break;
      }
    default:
      {
	write_state = 0;
	break;
      }
    } /* switch */

    buf++;
  }

  return count;
}

#define GPIO4 (1<<3)

void put_light_snom_display(int bOn)
{
  if(bOn)
    set_port(PORT_SNOM_CSR, GPIO4, 0);
  else
    set_port(PORT_SNOM_CSR, 0, GPIO4);
}

#define GPIO5 (1<<2)   /* ~CE - chip enable */
#define GPIO6 (1<<1)   /* DOWN/~UP - direction of level */
#define GPIO7 (1<<0)   /* PULSE - falling edge triggers the pulse */

void inc_contrast_snom_display(int bUp)
{
/* CE low, set dir, set pulse high */
    set_port(PORT_SNOM_CSR, (!bUp ? GPIO6 : 0) | GPIO7, GPIO5 | (bUp ? GPIO6 : 0));
/* Pulse */
    set_port(PORT_SNOM_CSR, 0, GPIO7);
    set_port(PORT_SNOM_CSR, GPIO7, 0);
/* CE hi */
    set_port(PORT_SNOM_CSR, GPIO5, 0);
}

int init_snom_display(void)
{
	write_state = 0;

	ks0108b_init();

	put_light_snom_display(1); // let there be light
	inc_contrast_snom_display(1); // put contrast in defined state
	
	return 0;
}

#endif




