/*
 * Support for ICE (currently at http://www.ietf.org/internet-drafts/draft-ietf-mmusic-ice-00.txt)
 *
 * Author: C. Stredicke, snom technology AG
 */


#include <linux/socket.h>
#include <linux/in.h>
#include "snomaud.h"

#define MAPPED_ADDRESS 0x0001
#define RESPONSE_ADDRESS 0x0002
#define STUN_FLAGS 0x0003
#define SOURCE_ADDRESS 0x0004
#define CHANGED_ADDRESS 0x0005
#define CMS_SIGNED_DATA 0x0006

#define BODY_BEGIN 20

// make this thing known:
// struct socket;

extern struct socket *rtp_socket[MAX_NUM_RTP];

int snom_aud_set_destination(struct socket *rtp_socket, struct in_addr *addr);
int snom_aud_set_default_destination(int c);
void snom_aud_ice_result(int c, struct in_addr *adr, int port);
void snom_aud_sendmsg(struct socket *skt, char *data, int len, int dport);

// little dirty helper routine:
static void conv_bytes2adr(struct in_addr *adr, unsigned char *response_adr) {
  unsigned char *a = (unsigned char*)adr;
  *a++ = response_adr[0];
  *a++ = response_adr[1];
  *a++ = response_adr[2];
  *a++ = response_adr[3];
}

// little dirty helper routine:
#if 0
static void conv_adr2bytes(unsigned char *response_adr, struct in_addr *adr) {
  unsigned char *a = (unsigned char*)adr;
  response_adr[0] = *a++;
  response_adr[1] = *a++;
  response_adr[2] = *a++;
  response_adr[3] = *a++;
}
#endif

// get the message type:
static inline int get_short(const char *msg, int offset) {
  const unsigned char c1 = (unsigned char)(msg[offset]);
  const unsigned char c2 = (unsigned char)(msg[offset + 1]);
  return ((int)c1 << 8) | (int)c2;
}

// get the message length:
static inline unsigned long get_long(const char *msg, int offset) {
  const unsigned char c1 = (unsigned char)(msg[offset]);
  const unsigned char c2 = (unsigned char)(msg[offset + 1]);
  const unsigned char c3 = (unsigned char)(msg[offset + 2]);
  const unsigned char c4 = (unsigned char)(msg[offset + 3]);
  return ((unsigned long)c1 << 24) | 
    ((unsigned long)c2 << 16) | 
    ((unsigned long)c3 << 8) | 
    (unsigned long)c4;
}

static inline void add_attribute(unsigned char *rest, int *rest_length, unsigned char *attribute, int attribute_length, int code) {
  rest += *rest_length;
  *rest++ = (unsigned char)((code >> 8) & 0xff);
  *rest++ = (unsigned char)(code & 0xff);
  *rest++ = (unsigned char)((attribute_length >> 8) & 0xff);
  *rest++ = (unsigned char)(attribute_length & 0xff);
  *rest_length += 4 + attribute_length;
  while(attribute_length-- > 0) *rest++ = *attribute++;
}

static inline void get_address(unsigned char *rest, const unsigned char *adr, int port) {
  *rest++ = (unsigned char)0; // n.c.
  *rest++ = (unsigned char)0x01; // IPv4
  *rest++ = (unsigned char)((port >> 8) & 0xff);
  *rest++ = (unsigned char)(port & 0xff);
  *rest++ = adr[0];
  *rest++ = adr[1];
  *rest++ = adr[2];
  *rest++ = adr[3];
}

/* 
 * handle a STUN pdu:
 * msg points to the message, length is the length of the message
 * original_addr points to the 4 bytes of the address, the port is the port number
 */
void snom_on_stun_available(int c,
			    struct ice_contact *ice_contacts, int ice_contacts_length,
			    unsigned char *msg, int msg_length, const unsigned char *original_addr, int original_port) {
  // get the origin address in str notation:
  int length;
  unsigned char *rest;
  int msg_type;
  int rest_length;
  int change_ip = 0;
  int change_port = 0;
  int has_mapped_adr = 0;
  static unsigned char response[128];
  int response_length;
  unsigned char response_adr[4];
  int response_port;
  int j;

  msg_type = get_short(msg, 0);
  if(msg_type == 0x0101) {
    int i, j;
    for(i = 0; i < ice_contacts_length; i++) {
      int equal = 1;
      for(j = 0; j < 16; j++) {
	if(msg[4 + j] != ice_contacts[i].id[j]) {
	  equal = 0;
	  break;
	}
      }
      if(equal) {
        // ok we have an ICE answer!!!
	struct in_addr adr;
	adr.s_addr = ice_contacts[i].dest_adr;
        snom_aud_ice_result(c, &adr, ice_contacts[i].dest_port);
        break;
      }
    }
    return;
  }
  else if(msg_type != 0x0001) {
    // Discarding unknown STUN type
    return;
  }
  
  // check that the length is ok:
  length = get_short(msg, 2);
  if(msg_length - length != BODY_BEGIN) return; // Wrong length

  // set the default answer address:
  response_adr[0] = original_addr[0];
  response_adr[1] = original_addr[1];
  response_adr[2] = original_addr[2];
  response_adr[3] = original_addr[3];
  response_port = original_port;

  rest = &msg[BODY_BEGIN];
  rest_length = msg_length - BODY_BEGIN;

  for(j = 0; j < BODY_BEGIN; j++) response[j] = msg[j];
  response_length = BODY_BEGIN;

  while(rest_length > 4) {
    int att_type = get_short(rest, 0);
    int att_length = get_short(rest, 2);
    unsigned char *att_content = &rest[4];
    if(att_length + 4 > rest_length) break; // avoid segmentation fault
    rest = &rest[att_length + 4];
    rest_length -= att_length + 4;

    switch(att_type) {
    case 0x0001: // MAPPED-ADDRESS
      if(response_length + att_length > sizeof(response)) return; // avoid segmentation violation
      add_attribute(response, &response_length, att_content, att_length, 0x0001);
      has_mapped_adr = 1;
      break;
      
    case 0x0002: // RESPONSE-ADDRESS
      if(att_length >= 8) {
	const char family = att_content[1];
	if(family == 0x01) {
	  // IPv4
	  response_adr[0] = att_content[4];
	  response_adr[1] = att_content[5];
	  response_adr[2] = att_content[6];
	  response_adr[3] = att_content[7];
	  response_port = get_short(att_content, 2);
	}
	else {
	  // Unknown STUN family
	  return;
	}
      }
      break;

    case 0x0003: // FLAGS
      if(att_length >= 4) {
	const char c1 = att_content[3];
	if(c1 & 0x4) change_ip = 1; // thanks for an unclear specification
	if(c1 & 0x2) change_port = 1; // thanks for an unclear specification
      }
      break;

    default:
      if(att_type < 0x8000) {
        // Ignoring packet because of attribute type
	return;
      }
      break;
    }
  } // while
  
  // add mapped address
  if(!has_mapped_adr) {
    unsigned char mapped_address[8];
    get_address(mapped_address, original_addr, original_port);
    add_attribute(response, &response_length, mapped_address, 8, 0x0001);
  }
  
  // set the length of the packet
  response[0] = (unsigned char)0x01; // response
  response[2] = (unsigned char)(((response_length - BODY_BEGIN) >> 8) & 0xff);
  response[3] = (unsigned char)((response_length - BODY_BEGIN) & 0xff);
  
  // send the response:
  if(!change_ip && !change_port) {
    struct in_addr adr;
    conv_bytes2adr(&adr, response_adr);
    if(snom_aud_set_destination(rtp_socket[c], &adr)) {
      snom_aud_sendmsg(rtp_socket[c], response, response_length, response_port);
    }
    snom_aud_set_default_destination(c);
  }
}

#if 0
// get a STUN packet with a given ID (total 28 bytes)
static void get_stun_packet(unsigned char *request, int *length, unsigned char *id) {
  int i;
  *request++ = (unsigned char)0;
  *request++ = (unsigned char)0x01;
  *request++ = (unsigned char)0; // length MSB
  *request++ = (unsigned char)8; // length LSB

  // transaction id:
  for(i = 0; i < 16; i++) *request++ = *id++;
  
  *request++ = (unsigned char)((STUN_FLAGS >> 8) & 0xff);
  *request++ = (unsigned char)(STUN_FLAGS & 0xff);
  *request++ = (unsigned char)((4 >> 8) & 0xff);
  *request++ = (unsigned char)(4 & 0xff);
  *request++ = 0;
  *request++ = 0;
  *request++ = 0;
  *request++ = 0;
  *length = 28;
}

// send out ICE measure packets:
void snom_do_ice(int c, struct ice_contact *ice_contacts, int ice_contacts_length) {
  int i;
  static unsigned char stun[128];
  for(i = 0; i < ice_contacts_length; i++) {
    struct in_addr adr;
    int length;
    get_stun_packet(stun, &length, ice_contacts[i].id);

    adr.s_addr = ice_contacts[i].dest_adr;
    if(snom_aud_set_destination(rtp_socket[c], &adr)) {
      snom_aud_sendmsg(rtp_socket[c], stun, length, ice_contacts[i].dest_port);
    }
  }
  snom_aud_set_default_destination(c);
}
#endif
