/*
 *  linux/drivers/char/snomflash.c
 *
 *  Copyright (C) 2000, snom technology AG
 *
 *  Author: Nicolas-Peter Pohland
 *
 *  This driver is for the Fujitsu flash on the SNOM board.
 *  It will implement (in this order) :
 *  1. The read() and write() system calls.
 *  2. ioctl() can be used to erase the flash. 
 *  3. seek() is implemented.
 *  4. 1x2MB, 2x2MB and 1x4MB configuration support
 *  5. Settings support has been added
 *  6. Setting the serial number has been added.
 *
 */

/***********************************************************************

************************************************************************/



#include <linux/config.h>

#ifdef CONFIG_SNOM_FLASH

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>

#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/capability.h>
#include <linux/delay.h>

#include <asm/mpc8xx.h>

#ifdef  CONFIG_SNOM_DISP
#ifndef  CONFIG_SNOM200
#include "../../arch/ppc/8xx_io/progress.h"
#endif
#endif

#include "../../arch/ppc/8xx_io/commproc.h"

/* Constants */

/* Debug */

#undef DEBUG 

#if DEBUG
#define DEB(x)        printk(x);
#define DEB2(x,y)     printk(x,y);
#define DEB3(x,y,z)   printk(x,y,z);
#define DEB4(x,y,z,a) printk(x,y,z,a);
#else
#define DEB(x)
#define DEB2(x,y)
#define DEB3(x,y,z)
#define DEB4(x,y,z,a)
#endif

static int snom_flash_type	= 0;
static int no_of_chips 		= 0;

static unsigned int serial_number_adr    = 0;
static unsigned int firmware_sector_adr  = 0;
static unsigned int settings_sector_adr  = 0;
static unsigned int flash_size	= 0;
static unsigned int flash_size_per_chip	= 0;

#define NO_OF_SECTORS_2MB 35

static unsigned int sector_adr_2MB[NO_OF_SECTORS_2MB] = {
0x00000, 0x02000, 0x03000, 0x04000, 0x08000, 0x10000, 0x18000,
0x20000, 0x28000, 0x30000, 0x38000, 0x40000, 0x48000, 0x50000,
0x58000, 0x60000, 0x68000, 0x70000, 0x78000, 0x80000, 0x88000,
0x90000, 0x98000, 0xa0000, 0xa8000, 0xb0000, 0xb8000, 0xc0000,
0xc8000, 0xd0000, 0xd8000, 0xe0000, 0xe8000, 0xf0000, 0xf8000
}; 

#define NO_OF_SECTORS_4MB 71


// sector address table for bottom up flash
static unsigned int sector_adr_4MB[NO_OF_SECTORS_4MB] = {
0x000000, 0x001000, 0x002000, 0x003000, 0x004000, 0x005000, 0x006000, 0x007000,
0x008000, 0x010000, 0x018000, 0x020000, 0x028000, 0x030000, 0x038000, 0x040000,
0x048000, 0x050000, 0x058000, 0x060000, 0x068000, 0x070000, 0x078000, 0x080000,
0x088000, 0x090000, 0x098000, 0x0a0000, 0x0a8000, 0x0b0000, 0x0b8000, 0x0c0000, 
0x0c8000, 0x0d0000, 0x0d8000, 0x0e0000, 0x0e8000, 0x0f0000, 0x0f8000, 0x100000,
0x108000, 0x110000, 0x118000, 0x120000, 0x128000, 0x130000, 0x138000, 0x140000,
0x148000, 0x150000, 0x158000, 0x160000, 0x168000, 0x170000, 0x178000, 0x180000,
0x188000, 0x190000, 0x198000, 0x1a0000, 0x1a8000, 0x1b0000, 0x1b8000, 0x1c0000, 
0x1c8000, 0x1d0000, 0x1d8000, 0x1e0000, 0x1e8000, 0x1f0000, 0x1f8000
}; 

int read_word(int nIdx, int nAdr)
{
  if(nIdx == 0)
  	return (int) (*((unsigned short *) FLASH_BASE_ADDR + nAdr));
  else {
  	return (int) (*((unsigned short *) FLASH_BASE_PLUS_2MB_ADDR + nAdr));
  }
}

void write_word(int nIdx, int nAdr, int nVal)
{
  if(nIdx == 0) {
  	*((unsigned short *) FLASH_BASE_ADDR + nAdr) = (unsigned short) nVal;
  }else {
  	*((unsigned short *) FLASH_BASE_PLUS_2MB_ADDR + nAdr) = (unsigned short) nVal;
  }
}

/****** Erase flash */

static int erase_snom_flash(void)
{
  unsigned long end_jiffies;
  int i, j, nVal, nError = 0;
  /* Send out the chip erase sequence */

  for(j=no_of_chips-1; j >= 0; j--) {

  	printk("Erasing flash %d\n", j+1);

  	write_word(j, 0x000, 0xF0);

     	write_word(j, 0x555, 0xAA);
     	write_word(j, 0x2AA, 0x55);
     	write_word(j, 0x555, 0x80);
     	write_word(j, 0x555, 0xAA);
     	write_word(j, 0x2AA, 0x55);

	write_word(j, 0x555, 0x10);
	
  	/* Prepare to latch data in */

	for(i=0;i<180;i++) {
		nVal = read_word(j, 0);
		if(nVal == 0xffff) break;
  		printk(".");

      end_jiffies = jiffies + 200;
      while(jiffies < end_jiffies)
	schedule();
	}

	if(nVal != 0xffff) {
  		nError = -EIO;
	}

  	if(nError)
  		printk("failed.\n");
  	else
  		printk("done.\n");
  }

  return nError;
}

int erase_snom_flash_sectors(int uiFrom, int uiTo)
{
  unsigned long end_jiffies;
  unsigned int nAdr;
  unsigned int no_of_sectors, *puiSectors=NULL;
  int i, j, nIdx, nVal, nError = 0;

  DEB3("snomflash: erase from %08x to %08x.\n", uiFrom, uiTo);

  puiSectors = (snom_flash_type == FLASH_4MB) ?
    sector_adr_4MB : sector_adr_2MB;

  no_of_sectors = (snom_flash_type == FLASH_4MB) ? 
    NO_OF_SECTORS_4MB : NO_OF_SECTORS_2MB;

  for(nIdx=0;nIdx < no_of_chips;nIdx++) {

    for(j=0;j < no_of_sectors;j++) {

      nAdr = puiSectors[j] + ((nIdx * flash_size_per_chip) >> 1);

      if((nAdr >= uiFrom) && (nAdr < uiTo)) {

	if(nIdx) {
	  nAdr %= (flash_size_per_chip >> 1);
	}  

	write_word(nIdx, 0x000, 0xF0);

	write_word(nIdx, 0x555, 0xAA);
	write_word(nIdx, 0x2AA, 0x55);
	write_word(nIdx, 0x555, 0x80);
	write_word(nIdx, 0x555, 0xAA);
	write_word(nIdx, 0x2AA, 0x55);
	
	write_word(nIdx, nAdr, 0x30);
    
	/* Prepare to latch data in */

	for(i=0;i<20;i++) {
	  nVal = read_word(nIdx, nAdr);
	  if(nVal == 0xffff) break;
	  //printk(".");
	  
	  end_jiffies = jiffies + 100; // in 2 secs
	  while(jiffies < end_jiffies)
	    schedule();
	}
	
	if(nVal != 0xffff) {
	  nError = -EIO;
	}
      }
    }
  }
      
  /*if(nError)
    printk("failed.\n");
  else
    printk("done.\n");*/
    
  return nError;
}

int erase_snom_flash_boot_sector(void)
{
  DEB3("snomflash: erase boot sector %08x to %08x.\n", 0, 
        (serial_number_adr-1) >> 1);

  return erase_snom_flash_sectors(0, (serial_number_adr-1) >> 1);
}

int erase_snom_flash_firmware_sector(void)
{
  DEB3("snomflash: erase from %08x to %08x.\n", firmware_sector_adr >> 1, 
      (settings_sector_adr-1) >> 1);

  return erase_snom_flash_sectors(
    	 	  firmware_sector_adr >> 1, (settings_sector_adr-1) >> 1);
}

int erase_snom_flash_settings_sector(void)
{
  DEB3("snomflash: erase settings secotr %08x to %08x.\n", 
      settings_sector_adr >> 1, (flash_size-1) >> 1);

  return erase_snom_flash_sectors(settings_sector_adr >> 1, 
				  (flash_size-1) >> 1);
}

int erase_snom_flash_serial_number(void)
{
  return erase_snom_flash_sectors(
                  serial_number_adr >> 1, (serial_number_adr + 16) >> 1);
}

/****** Generic driver stuff. */

static struct file_operations snom_flash_fops;
static int                    snom_flash_used = 0;

static ssize_t write_snom_flash(struct file * file, const char * buf,
			  size_t count, loff_t *ppos)
{
  int nVal, i, j, nIdx;
  unsigned long p = *ppos;
  unsigned short sVal;

  DEB3("snomflash: write - count %u offset %08x\n", (unsigned) count, 
       (unsigned) p);

  if(count < 0)
    return -EINVAL;

  if(p >= flash_size)
    return -ENOSPC;

  if(p + count >= flash_size)
    count = flash_size - 1 - p;

  if(count <= 0)
	return 0;

  count >>= 1; /* no odd byte number */
  count <<= 1;
  
  /* Reset to read */
   write_word(0, 0x0000, 0xF0);
   if(no_of_chips == 2)
   	write_word(1, 0x0000, 0xF0);

  for(i=0; i < count; i+= 2) {

    nIdx = ((no_of_chips == 1) || (p < FLASH_2MB_SIZE)) ? 0 : 1;

    write_word(nIdx, 0x555, 0xAA);
    write_word(nIdx, 0x2AA, 0x55);
    write_word(nIdx, 0x555, 0xA0);

    if (copy_from_user(&sVal, &buf[i], 2)) 
      return -EFAULT;

    write_word(nIdx, (p % flash_size_per_chip) >> 1, sVal);

    j=0;

    do {
      nVal = read_word(nIdx,(p % flash_size_per_chip) >> 1);
      
    } while((sVal != nVal) && (++j < 1000000));

#if DEBUG
    if(j >= 1000000)
      DEB4("snomflash: write failed at %08x, %08x != %08x\n", (unsigned) p,
	  (unsigned) sVal, (unsigned) nVal);
#endif

    p += 2;
  }

  *ppos += count;

  return (ssize_t) count;
}

static ssize_t read_snom_flash(struct file * file, char * buf,
			       size_t count, loff_t *ppos)
{
  unsigned long p = *ppos;
  ssize_t read = 0;

  if(count < 0)
    return -EINVAL;

  if(p >= flash_size)
    return -1;

  if(p + count >= flash_size)
    count = flash_size - 1 - p;
  
  if(count == 0)
    return -1; /* end the suffering ... */

  copy_to_user(buf, (void *) (p + FLASH_BASE_ADDR), count);

  read += count;
  *ppos += read;

  return read;
}

/*
 By convention the last sector is used for system settings.
	
 The system settings are the boot command line that is parsed by net/ipv4/config.c
 essentially to set and store the values needed by the system before a server
 can be found.

 In order to be able to burn these settings the seek is not used in 
 the conventional way. Instead, values are used to indicate whether the system,
 serial number or the first sector is of interest.

*/

static loff_t lseek_snom_flash(struct file * file, loff_t offset, int orig)
{

  switch (orig) {
  case SNOM_FLASH_SETTINGS_SECTOR:

  DEB("snomflash: lseek settings\n");

    file->f_pos = settings_sector_adr; break;

  case SNOM_FLASH_SEEK_CUR:

  DEB3("snomflash: lseek seek current offset = %x(%d)\n", 
       (unsigned) offset, (unsigned) offset);

    file->f_pos += offset; break;
  case SNOM_FLASH_BEGIN:
  default:

  DEB("snomflash: lseek begin\n");

    file->f_pos = 0;
  }
  return file->f_pos;
}

static int release_snom_flash(struct inode *inode, struct file * filp) 
{
  DEB("snomflash: release\n"); 

  snom_flash_used = 0;  return 0;
}

static int ioctl_snom_flash(struct inode *inode, struct file *file,
			    unsigned int cmd, unsigned long arg)
{
  int nError;

  DEB("snomflash: ioctl -");

  switch(cmd) {
  case SNOM_FLASH_ERASE_SETTINGS_SECTOR:
    {
      DEB("erase settings sector");
      file->f_pos = settings_sector_adr; 
      nError = erase_snom_flash_settings_sector();
      break;
    }
  case SNOM_FLASH_ERASE_FIRMWARE_SECTOR:
    {
      DEB("erase settings sector");
      file->f_pos = firmware_sector_adr; 
      nError = erase_snom_flash_firmware_sector();
      break;
    }
  case SNOM_FLASH_ERASE_BOOT_SECTOR:
    {
      DEB("erase boot sector");
      file->f_pos = 0; 
      nError = erase_snom_flash_boot_sector();
      break;
    }
  case SNOM_FLASH_ERASE_SERIAL_NUMBER:
    {
      DEB("erase serial number");
      file->f_pos = serial_number_adr; 
      nError = erase_snom_flash_serial_number();
      break;
    }
  case SNOM_FLASH_ERASE:
    {
      DEB("erase flash");
      file->f_pos = 0; 
      nError = erase_snom_flash();
      break;
    }
  default:
      DEB("illegal cmd\n");
    return -EINVAL;
  }

  DEB("\n");
  
  return nError;
}

static int open_snom_flash(struct inode * inode, struct file * filp)
{
  DEB("snomflash: open\n");

  if(snom_flash_used)
    return -EBUSY;

  switch (MINOR(inode->i_rdev)) {
  case 1:

    filp->f_op = &snom_flash_fops;
    break;
  default:
    return -ENXIO;
  }

  snom_flash_used = 1;

  return 0;
}

static struct file_operations snom_flash_fops = {
  llseek:            lseek_snom_flash,
  read:              read_snom_flash,
  write:             write_snom_flash,
  ioctl:             ioctl_snom_flash,		
  open:              open_snom_flash,
  release:           release_snom_flash		
};

void snom_flash_init(void)
{
  int serial_num;
  unsigned char *puc = (unsigned char *) 0xfe030003;
  serial_num = *puc;
  serial_num <<= 8;
  puc++;
  serial_num |= *puc;
  serial_num <<= 8;
  puc++;
  serial_num |= *puc;          

  if(serial_num >= 0 && serial_num < 1000)
	snom_flash_type = FLASH_2MB;
  else if(serial_num >= 1000 && serial_num < 2000) 
	snom_flash_type = FLASH_2x2MB;
  else 
	snom_flash_type = FLASH_4MB;

  no_of_chips   = (snom_flash_type != FLASH_2x2MB) ? 1 : 2;
  flash_size = (snom_flash_type == FLASH_2MB) ? FLASH_2MB_SIZE : FLASH_4MB_SIZE;
  flash_size_per_chip = (snom_flash_type == FLASH_4MB) ? FLASH_4MB_SIZE : FLASH_2MB_SIZE;

  printk("SNOM Flash Driver $Revision: 1.3 $ - 2.4.18\n");

  switch(snom_flash_type) {
  case FLASH_2MB:
    printk("Assuming 1x2MB flash.\n");
    settings_sector_adr = sector_adr_2MB[NO_OF_SECTORS_2MB-1];
    break;
  case FLASH_2x2MB:
    printk("Assuming 2x2MB flash.\n");
    settings_sector_adr = sector_adr_2MB[NO_OF_SECTORS_2MB-1] + 
      (FLASH_2MB_SIZE >> 1);
    break;
  case FLASH_4MB:
    printk("Assuming 1x4MB flash.\n"); 
    settings_sector_adr = sector_adr_4MB[NO_OF_SECTORS_4MB-1];

    DEB2("settings_sector_adr = %x\n", settings_sector_adr);

    break;
  default:
    printk("Unknown configuration\n"); break;
  }

  serial_number_adr = 0x00030000;
  settings_sector_adr <<= 1;
  firmware_sector_adr = (256 *1024);

  if (register_chrdev(SNOM_FLASH_MAJOR,"snomflash",&snom_flash_fops))
    printk("unable to get major %d for flash devs\n", 
	   SNOM_FLASH_MAJOR);

#ifdef CONFIG_SNOM_DISP
#ifndef CONFIG_SNOM200
        write_snom_display(progress_pic[7], sizeof(progress_pic[0]));
#else
        progress_snom_display();
#endif
#endif
}

#endif

