/***********************************************************************
 *
 *  linux/drivers/char/snomkeyb.c
 *
 *  Copyright (C) 2000, snom technology AG
 *
 *  This driver is for the keyboard module connected to the SNOM board.
 *
 *  09.02.01 Added support for new matrix keyboard and MSB encodes whether
 *           key is being released.
 *
 *  30.05.00 Updated active polling for driver to support blocking, i.e.
 *           select.
 *
 *  04.09.00 Extended driver to use a matrix keyboard via port A.
 *
 ***********************************************************************/

/***********************************************************************

 This driver is used to drive the 74166 serial latches for the keyboard.
 Well, this orginally was true - now we have a EPM3064-10.

 The data lines are expected to be connected as follows:

 1  KBO0  - PA0   ~PE - Parallel enable
 2  KBO1  - PA1    CP - Clock 
 3  KBI0  - PA3    Q7 - Serial data

 The old matrix keyboard has the following pinout:

 PA4, PA6, PA7, PA8, PA9, PA12 are OUT[0-5]

 The keyboard has to be patched:
    a. Removed the diodes 
    b. Short circuit D7
    c. Disconnect the other side (not connected with D7) of pad 20
       (lower/left) and connect with pin 9
    d. Connect D8 with pin 11 
    e. Connect pins 5,6,7,8,9,10 with OUT[0-5]
    f. Connect pins 1,2,3,4 and 11 with pins 9,10,11,12 and 13 of
       CON4 (keyboard) of the main board.

************************************************************************/

#include <linux/config.h>

#if defined(CONFIG_SNOM_KEYB)

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>

#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/capability.h>
#include <linux/poll.h>
#include <linux/timer.h>

#include <asm/port.h>

#ifdef  CONFIG_SNOM_DISP
#include "../arch/ppc/8xx_io/progress.h"
#endif


/* Constants */

#define KEYB_BUFFER_SIZE 16

#define DELAY		/* for(j=0;j<100;j++); */


/* The data bus and these lines are located on port B. */

#define KBO0_A             (1<<15)
#define KBO1_A             (1<<14)
#define KBI0_A             (1<<12)

/* Hook is on port C pin 14 */

#define HOOK_PIN           (1<<1) 

#define OUT0 (1<<11)
#define OUT1 (1<<9)
#define OUT2 (1<<8)
#define OUT3 (1<<7)
#define OUT4 (1<<6)
#define OUT5 (1<<3)

#define OUT_MASK (OUT0 | OUT1 | OUT2 | OUT3 | OUT4 | OUT5)

#define IN0 0x80
#define IN1 0x40
#define IN2 0x20
#define IN3 0x10
#define IN4 0x08

/* Debug */

#if 0
#define DEB(x)          printk(x);
#define DEB2(x,y)       printk(x,y);
#define DEB3(x,y,z)     printk(x,y,z);
#define DEB4(x,y,z,a)   printk(x,y,z,a);
#define DEB5(x,y,z,a,b) printk(x,y,z,a,b);
#else
#define DEB(x)
#define DEB2(x,y)
#define DEB3(x,y,z) 
#define DEB4(x,y,z,a)
#define DEB5(x,y,z,a,b) 
#endif

static struct timer_list rt_periodic_timer = {
  { NULL, NULL },
  0L,
  0L,
  NULL
};

static char keyb_buffer[KEYB_BUFFER_SIZE];
static int  keyb_read_idx = 0;
static int  keyb_write_idx = 0;
static int  keyb_type = 0;

DECLARE_WAIT_QUEUE_HEAD(proc_list);

static void keyb_init(void)
{
  /*** Set the ports to the correct mode. */

  set_port_dir(PORT_A, OUTPUT_PINS, KBO0_A | KBO1_A);
  set_port_dir(PORT_A, INPUT_PINS,  KBI0_A);

  set_port_dir(PORT_C, INPUT_PINS,  HOOK_PIN);

  if (keyb_type == 2) {
    set_port_dir(PORT_A, OUTPUT_PINS, OUT_MASK);
  }
}

static int keyb_read_bits(void) {

    int i, nValue = 0, nPortVal;

    /* Set PE and CP to Low */
    set_port(PORT_A, 0, KBO0_A | KBO1_A);
    DELAY;
    /* CP to High to latch in data */
    set_port(PORT_A, KBO1_A, 0);
    DELAY;
    /* CP to Low and PE to High */
    set_port(PORT_A, KBO0_A, KBO1_A);
    DELAY;

    for(i=0; i<16;i++) {
      /* Read data */
      nPortVal = read_port(PORT_A);
      /* Zero means pressed. */
      if((nPortVal & KBI0_A) != 0) {
        nValue |= (1<<i);
      }

      set_port(PORT_A, KBO1_A, 0);
      DELAY;
      set_port(PORT_A, 0, KBO1_A);
      DELAY;
    }

    DEB2(">%x - ", nValue);

    return nValue;
}


#define MIN_PAUSE 10   /* wait a 1/10th second */

static int keyb_getkey(void)
{
  static int nLastValue = -1;
  static int nPauseVal = 0; /* number of times no char was read. */
  /* Is used to suppress that keys come too fast. */
  int nValue, nPortVal;

  if(keyb_type == 0)
    return -1;

  nPauseVal++;

  if(keyb_type == 1) {

    nValue = keyb_read_bits();

    /* Key is only considered pressed if only it
     is pressed and it was not the same as the
     last time. */

    DEB3("> %x %x\n", nLastValue, nValue);

    switch(nValue) {
    case 0x8000:
      nValue = '0'; break;
    case 0x4000:
      nValue = '1'; break;
    case 0x2000:
      nValue = '2'; break;
    case 0x1000:
      nValue = '3'; break;
    case 0x0800:
      nValue = '4'; break;
    case 0x0400:
      nValue = '5'; break;
    case 0x0200:
      nValue = '6'; break;
    case 0x0100:
      nValue = '7'; break;
    case 0x0080:
      nValue = '8'; break;
    case 0x0040:
      nValue = '9'; break;
    case 0x0020:
      nValue = '#'; break;
    case 0x0010:
      nValue = '*'; break;
    case 0x0008:
      nValue = 'a'; break;
    case 0x0004:
      nValue = 'b'; break;
    case 0x0002:
      nValue = 'c'; break;
    case 0x0001:
      nValue = 'd'; break;
    case 0xc000:
      nValue = 'e'; break;
    case 0xa000:
      nValue = 'f'; break;
    case 0x9000:
      nValue = 'g'; break;
    case 0x8800:
      nValue = 'h'; break;
    default:
      nValue = -1;
    }
  } 
  else if (keyb_type == 2) {

    set_port(PORT_A, 0,    OUT_MASK);

    set_port(PORT_A, OUT0, 0);
    nPortVal = keyb_read_bits();
    if(nPortVal & IN0)
      nValue = 'e';
    else if(nPortVal & IN1)
      nValue = '4';
    else if(nPortVal & IN2)
      nValue = '8';
    else if(nPortVal & IN3)
      nValue = '#';
    else if(nPortVal & IN4)
      nValue = 'a';

    set_port(PORT_A, OUT1, OUT0);
    nPortVal = keyb_read_bits();
    if(nPortVal & IN0)
      nValue = '1';
    else if(nPortVal & IN1)
      nValue = '5';
    else if(nPortVal & IN2)
      nValue = '9';
    else if(nPortVal & IN4)
      nValue = 'b';

    set_port(PORT_A, OUT2, OUT1);
    nPortVal = keyb_read_bits();
    if(nPortVal & IN0)
      nValue = '2';
    else if(nPortVal & IN1)
      nValue = '6';
    else if(nPortVal & IN2)
      nValue = '0';
    else if(nPortVal & IN4)
      nValue = 'c';

    set_port(PORT_A, OUT3, OUT2);
    nPortVal = keyb_read_bits();
    if(nPortVal & IN0)
      nValue = '3';
    else if(nPortVal & IN1)
      nValue = '7';
    else if(nPortVal & IN2)
      nValue = '*';
    else if(nPortVal & IN4)
      nValue = 'd';

    set_port(PORT_A, OUT4, OUT3);
    nPortVal = keyb_read_bits();
    if(nPortVal & IN0)
      nValue = 'h';
    else if(nPortVal & IN1)
      nValue = 'f';

    set_port(PORT_A, OUT5, OUT4);
    nPortVal = keyb_read_bits();
    if(nPortVal & IN0)
      nValue = 'g';

    set_port(PORT_A, 0,    OUT_MASK);

  } else if(keyb_type == 3) {
	
    nValue = keyb_read_bits();

    /* Key is only considered pressed if only it
     is pressed and it was not the same as the
     last time. */

    DEB3("> %x %x\n", nLastValue, nValue);
	
    if(nValue & 0x8000)
    	nValue 	&= 0x03ff;
    else	
	nValue = -1;

    switch(nValue) {
    case 0x50:
      nValue = '0'; break;
    case 0x22:
      nValue = '1'; break;
    case 0x42:
      nValue = '2'; break;
    case 0x82:
      nValue = '3'; break;
    case 0x24:
      nValue = '4'; break;
    case 0x44:
      nValue = '5'; break;
    case 0x84:
      nValue = '6'; break;
    case 0x28:
      nValue = '7'; break;
    case 0x48:
      nValue = '8'; break;
    case 0x88:
      nValue = '9'; break;
    case 0x90:
      nValue = '*'; break;
    case 0x30:
      nValue = '#'; break;
    case 0x21:
      nValue = 'a'; break;
    case 0x41:
      nValue = 'b'; break;
    case 0x81:
      nValue = 'c'; break;
    case 0x101:
      nValue = 'd'; break;
    case 0x102:
      nValue = 'e'; break;
    case 0x104:
      nValue = 'f'; break;
    case 0x108:
      nValue = 'g'; break;
    case 0x110:
      nValue = 'h'; break;
    default:
      nValue = -1;
    }

    /* Indicate key released ...

    if(!keypressed) 
    {
	nValue = 0x80 | nLastValue;
	nLastValue = -1;
    } else {
    	nLastValue = nValue;
    }
    */
  } 

  /* A key is pressed: */
  if(nValue >= 0) {
    if ((nLastValue == nValue) || (nPauseVal < MIN_PAUSE)) {
      nValue = -1;
    } 
    else {
      nPauseVal = 0;
      nLastValue = nValue;
    }  
  } 

  /* No key pressed: */
  else {
    if(nPauseVal >= MIN_PAUSE) {
      if (nLastValue >= 0) {
	nValue = nLastValue + 128; /* key up event */
	nLastValue = -1; 
      }
    }
    else {
      nValue = -1;
    }
  }

  return nValue;
}

/* Returns the initial value of the hook and then
   when it changes. 'y' for hook on and 'z' hook off. */
static int keyb_get_hook(void)
{
  static int nLastValue = -1, nPauseVal = 0;
  int nValue = -1;

  if(keyb_type == 0)
    return -1;

  nValue = read_port(PORT_C) & HOOK_PIN ? 'z' : 'y';
  nPauseVal++;

  if(nValue >= 0) {
    if ((nLastValue == nValue) || (nPauseVal < MIN_PAUSE)) {
      nValue = -1;
    } else {
      nPauseVal = 0;
      nLastValue = nValue;
    }
  } 

  return nValue;
}

/****** Generic driver stuff. */

static struct file_operations snom_keyb_fops;

static ssize_t read_snom_keyb(struct file *file, char *buf, 
			      size_t count, loff_t *ppos)
{

  int i;
  int no_of_chars_available;

  DECLARE_WAITQUEUE(wait, current);

  DEB("snomkeyb: read being called.\n");
  
  if(count < 1 || buf == NULL) 
    return -EINVAL;

  /* Is buffer empty ? */
  if(keyb_write_idx == keyb_read_idx) {
    if(file->f_flags & O_NONBLOCK)
      return -EAGAIN;
    add_wait_queue(&proc_list, &wait);
  repeat_read:
    current->state = TASK_INTERRUPTIBLE;
    if((keyb_write_idx == keyb_read_idx) && !signal_pending(current)) {
      schedule();
      goto repeat_read;
    }
    current->state = TASK_RUNNING;
    remove_wait_queue(&proc_list, &wait);
  }
   
  if(signal_pending(current))
    return -ERESTARTSYS;

  no_of_chars_available = keyb_write_idx >= keyb_read_idx ?
    keyb_write_idx - keyb_read_idx :
    keyb_write_idx + KEYB_BUFFER_SIZE - keyb_read_idx;
  
  if(count > no_of_chars_available)
    count = no_of_chars_available;
    
  for(i=0; i<count;i++) {
    copy_to_user(&buf[i], &keyb_buffer[keyb_read_idx], 1);
    keyb_read_idx = ((keyb_read_idx + 1) % KEYB_BUFFER_SIZE);
  }

  return count;
}

static int open_snom_keyb(struct inode * inode, struct file * filp)
{
  int serial_num;
  unsigned char *puc = (unsigned char *) 0xfe030003;

  serial_num = *puc;
  serial_num <<= 8;
  puc++;
  serial_num |= *puc;
  serial_num <<= 8;
  puc++;
  serial_num |= *puc;

  DEB("snomkeyb: open being called.\n");

  switch (MINOR(inode->i_rdev)) {
  case 1:
    keyb_type = serial_num >= 30 ? 3 : 1; 
    break;
  case 2:
    keyb_type = 2;
    break;
  default:
    return -ENXIO;
  }

  filp->f_op = &snom_keyb_fops;	
  keyb_read_idx = 0;
  keyb_write_idx = 0;

  keyb_init();

  return 0;
}

static unsigned int poll_snom_keyb(struct file *filp, poll_table * wait)
{
  poll_wait(filp, &proc_list, wait);

  if(keyb_write_idx != keyb_read_idx) /* not empty */
    return POLLIN | POLLRDNORM;
  return 0;
}

static void rt_check_expire(unsigned long dummy)
{
  int nVal;
  unsigned long now = jiffies;

  /* Is there space to retrieve a char ? */
  
  if(((keyb_read_idx + 1) % KEYB_BUFFER_SIZE) != keyb_write_idx) {
    /* then try to get a char ... */

    nVal = keyb_getkey();

    if(nVal >= 0) {

      keyb_buffer[keyb_write_idx] = (char) nVal;
      keyb_write_idx = ((keyb_write_idx + 1) % KEYB_BUFFER_SIZE);

      wake_up_interruptible(&proc_list);
    }

    /* then try to get a char for hook status change ... */

    nVal = keyb_get_hook();

    if(nVal >= 0) {

      keyb_buffer[keyb_write_idx] = (char) nVal;
      keyb_write_idx = ((keyb_write_idx + 1) % KEYB_BUFFER_SIZE);

      wake_up_interruptible(&proc_list);
    }
  } /* else do nothing ... */

  rt_periodic_timer.expires = now + 1; /* Should be 100 HZ */
  add_timer(&rt_periodic_timer);
}

static struct file_operations snom_keyb_fops = {
  read:     read_snom_keyb,
  poll:     poll_snom_keyb,		
  open:     open_snom_keyb
};

void snom_keyb_init(void)
{
  printk("SNOM Keyboard $Revision: 1.2 $ $Date: 2002/12/14 17:02:03 $\n");

  rt_periodic_timer.function = rt_check_expire;

  rt_periodic_timer.expires = jiffies + 60; /* wait a few sec */
  add_timer(&rt_periodic_timer);
	
  if (register_chrdev(SNOM_KEYB_MAJOR,"kbd",&snom_keyb_fops))
    printk("unable to get major %d for memory devs\n", SNOM_KEYB_MAJOR);

#ifdef CONFIG_SNOM_DISP
  write_snom_display(progress_pic[6], sizeof(progress_pic[0]));
#endif
 
}

#endif




