/*****************************************************************************/
/*
 *      snomsnd.c  --  SNOM Technology audio driver.
 *
 *      Copyright (C) 1999  SNOM Technology AG
 *
 *	Author: Nicolas-Peter Pohland
 *
 *	Supported devices:
 *
 *	/dev/dsp    standard /dev/dsp device
 *
 *  	Revision history
 *
 *	09.11.99   0.1   Initial release
 *
 */
/*****************************************************************************/

#ifndef __KERNEL__
#  define __KERNEL__
#endif   
   
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/sound.h>
#include <linux/malloc.h>
#include <linux/soundcard.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/hardirq.h>

#include <asm/port.h>

/* --------------------------------------------------------------------- */

#define NUM(dev)    (MINOR(dev) & 0xf) /* low  nibble */

/* --------------------------------------------------------------------- */

#undef OSS_DOCUMENTED_MIXER_SEMANTICS
#define DBG(x) {}
/*#define DBG(x) {x}*/

/* --------------------------------------------------------------------- */

#define SNOM_MAGIC  4711 

// #define CONFIG_ISDNOVERLAN 1

/* --------------------------------------------------------------------- */

struct snom_snd_state {
  unsigned int  magic;           /* Magic number */
  int           dev_snd;       /* sound core needs this */
  int           fragsize;        /* the fragment size */
};

/* --------------------------------------------------------------------- */

static struct snom_snd_state *devs = NULL;

/* --------------------------------------------------------------------- */


/* --------------------------------------------------------------------- */

static const char invalid_magic[] = KERN_CRIT "snom snd: invalid magic value\n";

#define VALIDATE_STATE(s)                         \
({                                                \
	if (!(s) || (s)->magic != SNOM_MAGIC) { \
		printk(invalid_magic);            \
		return -ENXIO;                    \
	}                                         \
})

/* --------------------------------------------------------------------- */

DECLARE_WAIT_QUEUE_HEAD(snom_snd_proc_list);

/* --------------------------------------------------------------------- */

static loff_t snom_snd_llseek(struct file *file, loff_t offset, int origin)
{
  printk(KERN_INFO "snom_snd: seek\n");

  VALIDATE_STATE(devs);

  return -ESPIPE;
}

extern int snom_aud_write_avail(void *);
extern int snom_aud_read_avail(void *);

void snom_snd_callback(void)
{
  wake_up_interruptible(&snom_snd_proc_list);
}

extern void snom_aud_set_call_back(void (*fn)(void));

static unsigned int snom_snd_poll(struct file *file, poll_table * wait)
{
  int mask = 0;

  poll_wait(file, &snom_snd_proc_list, wait);

  if(snom_aud_write_avail(file->private_data) >= 320)
    mask |= POLLOUT | POLLWRNORM;

  if(snom_aud_read_avail(file->private_data) > 0)
    mask |= POLLIN | POLLRDNORM;

  return mask;
}

extern int snom_aud_write(void *v, const unsigned char *buf, int count);

static ssize_t snom_snd_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
  int ret_val;

  DECLARE_WAITQUEUE(write_wait, current);

#if 0
  printk(KERN_INFO "snom_snd: write\n");
#endif

  VALIDATE_STATE(devs);

  if(count < 1 || buffer == NULL) 
    return -EINVAL;

  if(snom_aud_write_avail(file->private_data) < 160) {
    
    if (file->f_flags & O_NONBLOCK) 
      return -EAGAIN;
    
    add_wait_queue(&snom_snd_proc_list, &write_wait);
  repeat_read:
    current->state = TASK_INTERRUPTIBLE;
    if((snom_aud_write_avail(file->private_data) < 160) && !signal_pending(current)) {
      schedule();
      goto repeat_read;
    }
    current->state = TASK_RUNNING;
    remove_wait_queue(&snom_snd_proc_list, &write_wait);
  }
  
  if(signal_pending(current))
    return -ERESTARTSYS;
  
  ret_val = snom_aud_write(file->private_data, buffer, count);

  return ret_val;
}

extern int  snom_aud_get_blocksize     (void *v);
extern int  snom_aud_get_caps          (void *v);
extern int  snom_aud_get_formats       (void *v);
extern void snom_aud_set_amp_gain      (int val);
extern void snom_aud_set_blocking_mode (void *v, int flags);
extern void snom_aud_set_filter1_level (int val);
extern void snom_aud_set_filter2_level (int val);
extern void snom_aud_set_filter3_level (int val);
extern void snom_aud_set_format        (void *v, int format);
extern void snom_aud_set_rx_gain       (int val);
extern void snom_aud_set_tx_gain       (int val);

extern void snom_aud_shutdown          (int val);
extern void snom_aud_shutdown_speaker  (int val);

extern void snom_aud_reset             (void *v);

static int snom_snd_ioctl(struct inode *inode, struct file *file,
   unsigned int cmd, unsigned long arg)
{
	VALIDATE_STATE(devs);

	switch (cmd) {
	case SNDCTL_DSP_SETDUPLEX:  // we are always duplex
	  {
	    return 0;
	  } 
	case SNDCTL_DSP_CHANNELS:   // we only support 1 channel  
	  {
	    return put_user(1, (int *)arg);
	  } 
	case SNDCTL_DSP_SPEED:
	  {
	    int speed;
	    get_user(speed, (int *)arg);
	    return speed == 8000 ? 0 : -EINVAL;
	  }
	case SNDCTL_DSP_STEREO:
	  {
	    put_user(-EFAULT, (int *) arg);
	    return -EFAULT;
	  }
	case SNDCTL_DSP_GETFMTS:
	  {
	    int format = snom_aud_get_formats(file->private_data);
	    return put_user(format, (int *)arg);		
	  }
	case SNDCTL_DSP_SETFMT: 
	  {
	    int format;
	    get_user(format, (int *)arg);
	    snom_aud_set_format(file->private_data, format);
	    return 0;
	  }
	case SNDCTL_DSP_GETCAPS:
	  {
	    int caps = snom_aud_get_caps(file->private_data);
	    return put_user(caps, (int *)arg);		
	  }
	case SNDCTL_DSP_NONBLOCK:
	  {
	    file->f_flags |= O_NONBLOCK;
	    snom_aud_set_blocking_mode(file->private_data, O_NONBLOCK);
	    return 0;
	  }
	case SNDCTL_DSP_GETBLKSIZE:
	  {
	    int blocksize = snom_aud_get_blocksize(file->private_data);
	    return put_user(blocksize, (int *)arg);
	  }
	case SNDCTL_DSP_SETFRAGMENT: // we have got an optimal value already.
	  {
	    return 0;
	  }
	case SNDCTL_DSP_RESET:
	  {
	    snom_aud_reset(file->private_data);
	    return 0;
	  }
	case SNDCTL_DSP_SHUTDOWN:
	  {
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_shutdown(val);
	    return 0;
	  }
	case SNDCTL_DSP_SHUTDOWN_SPEAKER:
	  {
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_shutdown_speaker(val);
	    return 0;
	  }
	case SNDCTL_DSP_SET_RX_GAIN_LEVEL:
	  {
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_set_rx_gain(val);
	    return 0;
	  }
	case SNDCTL_DSP_SET_AMP_GAIN_LEVEL:
	  {
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_set_amp_gain(val);
	    return 0;
	  }
	case SNDCTL_DSP_SET_TX_GAIN_LEVEL:
	  {
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_set_tx_gain(val);
	    return 0;
	  }
	case SNDCTL_DSP_SET_RTP_PARAMETER:
	  {
	    extern int snom_aud_set_rtp_parameters(struct snom_aud_rtp_parameter_s *parm);
	    struct snom_aud_rtp_parameter_s rtp_parm;
	    copy_from_user(&rtp_parm, (struct snom_aud_rtp_parameter_s *)arg, sizeof(rtp_parm));
	    return snom_aud_set_rtp_parameters(&rtp_parm);
	  }
	case SNDCTL_DSP_SET_RTP_PARAMETER2:
	  {
	    extern int snom_aud_set_rtp_parameters2(int channel, struct snom_aud_rtp_parameter_s *parm);
	    struct snom_aud_rtp_parameter2_s rtp_parm;
	    copy_from_user(&rtp_parm, (struct snom_aud_rtp_parameter2_s *)arg, sizeof(rtp_parm));
	    return snom_aud_set_rtp_parameters2(rtp_parm.channel, &rtp_parm.data);
	  }
	case SNDCTL_DSP_SET_ICE:
	  {
	    extern int snom_aud_set_ice_parameters(struct snom_aud_ice_parameter_s *parm);
	    struct snom_aud_ice_parameter_s ice_parm;
	    copy_from_user(&ice_parm, (struct snom_aud_ice_parameter_s *)arg, sizeof(ice_parm));
	    return snom_aud_set_ice_parameters(&ice_parm);
	  }
#ifdef CONFIG_ISDNOVERLAN
	case SNDCTL_DSP_SET_RPI_PARAMETER:
	  {
	    extern int snom_aud_set_rpi_parameters(struct snom_aud_rpi_parm_s *parm);
	    struct snom_aud_rpi_parm_s rpi_parm;
	    copy_from_user(&rpi_parm, (struct snom_aud_rpi_parm_s *)arg, sizeof(rpi_parm));
	    return snom_aud_set_rpi_parm(&rpi_parm);
	  }
#endif // CONFIG_ISDNOVERLAN
	case SNDCTL_DSP_SET_MIXER_MODE:
	  {
	    extern void snom_aud_set_mixer_mode(int mode);
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_set_mixer_mode(val);
	    return 0;
	  }
	case SNDCTL_DSP_SET_DTMF:
	  {
	    extern void snom_aud_set_dtmf(int dtmf);
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_set_dtmf(val);
	    return 0;
	  }
	case SNDCTL_DSP_SET_AUDIO:
	  {
	    extern void snom_aud_set_audio(unsigned int val);
	    unsigned int data;
	    get_user(data, (unsigned int *)arg);
	    snom_aud_set_audio(data);
	    return 0;
	  }
	case SNDCTL_DSP_SET_TOS:
	  {
	    extern void snom_aud_set_tos(int tos);
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_set_tos(val);
	    return 0;
	  }
	case SNDCTL_DSP_SET_RECORDING:
	  {
	    extern void snom_aud_start_recording(int on);
	    int val;
	    get_user(val, (int *)arg);
	    snom_aud_start_recording(val);
	    return 0;
	  }
#ifndef CONFIG_SNOM100
	case SNDCTL_DSP_GET_STAT:
	  {
	    extern int snom_aud_get_stat(struct snom_aud_rtp_stat_s *stat);
	    struct snom_aud_rtp_stat_s stat;
	    int rc;
	    if(copy_from_user(&stat, (struct snom_aud_rtp_stat_s *)arg, sizeof(stat)))
	      return -EFAULT;
	    rc = snom_aud_get_stat(&stat);
	    if(copy_to_user((void*)arg, &stat, sizeof(stat)))
	      return -EFAULT;
	    return rc;
	  }
#endif
	case SNDCTL_KB_READ_PORT:
	  {
	    extern int read_port(t_port port);
	    int rc, val;
	    get_user(val, (int *)arg);
            if (val < 0 || val > 3) return -EINVAL;
	    rc = read_port(val);
	    return rc;
	  }
	default:
	  printk(KERN_INFO "snom_snd: unknown ioctl %x\n", cmd);
	}
	return -EINVAL;
}

static int snom_snd_s_count = 0;

extern int snom_aud_open(int minor_num, void **, int mode, int flags);

static int snom_snd_open(struct inode *inode, struct file *filp)
{
  int ret_val;
  
#if 0
  printk(KERN_INFO "snom_snd: open\n");
#endif

  ret_val = snom_aud_open(MINOR(inode->i_rdev) - 3, &(filp->private_data),
     filp->f_mode, filp->f_flags);

  if(ret_val)
    return ret_val;

  snom_snd_s_count++;

  MOD_INC_USE_COUNT;
  return 0;
}

extern void snom_aud_close(void *);

static int snom_snd_release(struct inode *inode, struct file *filp)
{
#if 0
  printk(KERN_INFO "snom_snd: release\n");
#endif

  snom_aud_close(filp->private_data);
  filp->private_data = NULL;

  snom_snd_s_count--; /* release the device */
  MOD_DEC_USE_COUNT;
  return 0;
}

static /*const*/ struct file_operations snom_snd_fops = {
  llseek:   snom_snd_llseek,
  poll:     snom_snd_poll,
  write:    snom_snd_write,
  ioctl:    snom_snd_ioctl,
  open:     snom_snd_open,
  release:  snom_snd_release
};

/* --------------------------------------------------------------------- */

/* maximum number of devices */
#define NR_DEVICE 1

MODULE_AUTHOR("Nicolas-Peter Pohland, pohland@snom.de");
MODULE_DESCRIPTION("SNOM Snd Driver");

/* --------------------------------------------------------------------- */

int __init init_snom_snd(void)
{
  struct snom_snd_state *s; 

  printk(KERN_INFO "SNOM Snd Device, Copyright (C) 1999 SNOM Technology AG\n");
  
  if (!(s = kmalloc(sizeof(struct snom_snd_state), GFP_KERNEL))) {
    printk(KERN_WARNING "snom snd: out of memory\n");
    goto err_dev0;
  }
  memset(s, 0, sizeof(struct snom_snd_state));

  s->magic = SNOM_MAGIC;

  /* register devices */
  if ((s->dev_snd = register_sound_dsp(&snom_snd_fops, -1)) < 0)
  {
    printk(KERN_WARNING "snom snd: failed to register\n");
    goto err_dev1;
  }

  devs = s;

  snom_aud_set_call_back(snom_snd_callback);

  /* Success */
  return 0;

 err_dev1:
  kfree(s);
 err_dev0:  
  devs = NULL;
  return -ENODEV;
}

#ifdef EVER_NEEDED
static void __exit cleanup_snom(void)
{

  if(devs)
    {
      unregister_sound_dsp(devss->dev_snd);
      kfree(devss);
      devs=NULL;
    } 

  printk(KERN_INFO "snom: unloading\n");
}
#endif
